"""
Copyright (c) 2025 Brendan Meade. All rights reserved.

parsli: Earth data viewers
"""

from __future__ import annotations

import logging

__version__ = "0.10.1"
__all__ = ["__version__"]

# Set global config
logging.basicConfig(level=logging.CRITICAL)
